"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.KeyPair = void 0;
// a private and public key pair
const bsv_1 = require("bsv");
class KeyPair {
    constructor(priv, pub) {
        this.privateKey = priv;
        this.publicKey = pub;
    }
    static fromWif(wif) {
        const pk = bsv_1.PrivKey.fromWif(wif);
        return KeyPair.fromPrivateKey(pk);
    }
    static fromRandom() {
        const bsvkp = new bsv_1.KeyPair().fromRandom();
        return new KeyPair(bsvkp.privKey, bsvkp.pubKey);
    }
    static fromPrivateKey(pk) {
        const privateKey = pk;
        const publicKey = bsv_1.PubKey.fromPrivKey(pk);
        return new KeyPair(privateKey, publicKey);
    }
    toAddress() {
        return bsv_1.Address.fromPubKey(this.publicKey, 'livenet');
    }
    toBsvKeyPair() {
        return new bsv_1.KeyPair().fromPrivKey(this.privateKey);
    }
    toWif() {
        return this.privateKey.toWif();
    }
    toScript() {
        return this.toAddress().toTxOutScript();
    }
}
exports.KeyPair = KeyPair;
//# sourceMappingURL=KeyPair.js.map