"use strict";
// A Bitcoin Token envelope
// could be a utxo when unspent or could be a transaction input when getting spent
// has a status (unspent/spent)
// has a proof
// has list of transactions for inputs
Object.defineProperty(exports, "__esModule", { value: true });
exports.TokenEnvelope = void 0;
class TokenEnvelope {
    constructor(object) {
        // unspent or spent
        this.status = 'unspent';
        // utxo that holds these tokens when unspent
        this.utxo = null;
        // satoshi tokens in this envelope
        this.satoshis = 0;
        // merkle proof for these tokens when mined
        this.proof = null;
        // array of input transaction
        this.txns = [];
        if (object && object.status !== undefined)
            this.status = object.status;
        if (object && object.utxo !== undefined)
            this.utxo = object.utxo;
        if (object && object.satoshis !== undefined)
            this.satoshis = object.satoshis;
        if (object && object.proof !== undefined)
            this.proof = object.proof;
        if (object && object.txns !== undefined)
            this.txns = object.txns;
    }
    get isUnspent() { return this.status === 'unspent'; }
    markSpent() {
        if (!this.isUnspent)
            throw new Error(`already spent!`);
        this.status = 'spent';
    }
}
exports.TokenEnvelope = TokenEnvelope;
//# sourceMappingURL=TokenEnvelope.js.map