"use strict";
// validates transactions and proofs
// uses WOC proof format, should change to bsv proof standard format
Object.defineProperty(exports, "__esModule", { value: true });
exports.Validator = void 0;
const bsv_1 = require("bsv");
const IndexClient_1 = require("./IndexClient");
class Validator {
    constructor(index) {
        this.api = index || new IndexClient_1.IndexClient();
    }
    swapEndianness(buffer) {
        const result = Buffer.allocUnsafe(buffer.length);
        for (let i = 0, j = buffer.length - 1; i <= j; i++, j--) {
            result[i] = buffer[j];
            result[j] = buffer[i];
        }
        return result;
    }
    // tx is the tx to validate, proof is the proof for that tx
    validateProof(/*tx:any,*/ proof) {
        // calculate merkle proof
        const merkle = this.calculateMerkle(proof);
        // check if merkle is part of a block header
        return this.isMerkleInBlockHeader(merkle);
    }
    hash(left, right) {
        const catn8ted = Buffer.concat([this.swapEndianness(left), this.swapEndianness(right)]);
        //console.log('catn8ted', catn8ted)
        return this.swapEndianness(bsv_1.Hash.sha256Sha256(catn8ted));
    }
    calculateMerkle(proof) {
        //proof.hash is the tx hash
        let running_hash = Buffer.from(proof.hash, 'hex');
        proof.branches.forEach((branch) => {
            //console.log(branch)
            running_hash = (branch.pos === "R") ?
                this.hash(running_hash, Buffer.from(branch.hash, 'hex'))
                : this.hash(Buffer.from(branch.hash, 'hex'), running_hash);
        });
        return running_hash.toString('hex');
    }
    async isMerkleInBlockHeader(merkle) {
        //call api to see if merkle can be found
        const jblock = await this.api.getBlockHeaderForMerkleRoot(merkle);
        // {"merkleroot": "95a920b1002bed05379a0d2650bb13eb216138f28ee80172f4cf21048528dc60", "hash": "0000000000000000091216c46973d82db057a6f9911352892b7769ed517681c3"}
        const block = JSON.parse(jblock);
        return { isValid: block.merkleroot ? true : false, merkleroot: block["merkleroot"], blockhash: block["hash"] };
    }
}
exports.Validator = Validator;
//# sourceMappingURL=Validator.js.map