// holds a collection of token envelopes

import { TokenEnvelope } from "./TokenEnvelope"
export class Envelopes {
    _token_envelopes:TokenEnvelope[] = []

    // balance requires that _token_envelopes is a TokenEnvolope instance!
    get Balance():number {
        let bal = 0
        if (this._token_envelopes && this._token_envelopes.length>0) {
            this._token_envelopes.forEach(env => {
                if (env.isUnspent) {
                    bal += env.satoshis
                }
            })
        }
        return bal
    }

    get unspent() { 
        const u:any[] = []
        if (this._token_envelopes && this._token_envelopes.length>0) {
            this._token_envelopes.forEach(env => {
                if (env.isUnspent) {
                    u.push(env)
                }
            })
        }
        return u
    }

    //return utxos (unspent) from token envelopes
    get utxos() { 
        const u:any[] = []
        if (this._token_envelopes && this._token_envelopes.length>0) {
            this._token_envelopes.forEach(env => {
                if (env.isUnspent) {
                    u.push(env.utxo)
                }
            })
        }
        return u
    }

    //TODO: disable, load whole Envelopes instead
    set utxos(val:any[]) { 
        //TODO: merge utxos passed in with token_envelopes
        if (val) {
            val.forEach(u => {
                let envelope = this.findutxo(u)
                if (!envelope) {
                    envelope = new TokenEnvelope()
                    envelope.utxo = u
                    envelope.satoshis = u.value
                    this.add(envelope)
                }
            })
        }
    }

    clear() {this._token_envelopes = []}

    add (envelope: TokenEnvelope) {
        if (envelope) {
            // {"height":693098,"tx_pos":0,
            // "tx_hash":"3ab9cb22198b4de77dbd7848eb8795d269a5de36933a5ffe87ad6e167e144ad5",
            // "value":1000,
            // "script":"76a914a7e7b30ab0580dc8a090182d97903e056d4f080888ac"}
            if (!envelope.utxo.tx_hash ||!envelope.utxo.value) {
                throw new Error(`INVALID ENVELOPE utxo ${JSON.stringify(envelope.utxo)}`)
            }
            if (envelope.utxo && envelope.utxo.value) { envelope.satoshis = envelope.utxo.value }
            this._token_envelopes.push(envelope)
        } else console.log(`CANNOT LOAD ENVELOPE ${envelope}`)
    }

    load(envelopes:any) {
        this.clear()
        const whattoload = (envelopes._token_envelopes) ? envelopes._token_envelopes : envelopes
        for(let x=0;x<whattoload.length;x++) {
            const env = new TokenEnvelope(whattoload[x])
            this.add(env)
        }
    }

    // select utxos
    // {
    //     height: 693098,
    //     tx_pos: 0,
    //     tx_hash: '3ab9cb22198b4de77dbd7848eb8795d269a5de36933a5ffe87ad6e167e144ad5',
    //     value: 1000,
    //     script: '76a914a7e7b30ab0580dc8a090182d97903e056d4f080888ac'
    //   }
    selectUnspents(amount:number):any[] {
        // sort descending
        function compare(a:any, b:any) {          
            let comparison = 0;
            if (a.value < b.value) {
              comparison = 1;
            } else if (a.value > b.value) {
              comparison = -1;
            }
            return comparison;
          }
          const utxos = this.utxos
          utxos.sort(compare)
          const result:any[] = []
          let sum = 0
          utxos.forEach((utxo:any) => {
            if (sum < amount) {
                result.push(utxo)
                sum += utxo.value
            }
          })
          return result
    }

    findutxo(utxo:any):any {
        for(let x=0; x< this._token_envelopes.length; x++) {
            const env = this._token_envelopes[x]
            if (env.utxo.tx_hash === utxo.tx_hash && env.utxo.tx_pos === utxo.tx_pos) {
                return env
            }
        }
        return null
    }

    // mark the utxos as spent
    markSpent(utxos:any) {
        for( let x =0; x < utxos.length; x++) {
            const utxo = utxos[x]
            const envelope = this.findutxo(utxo)
            if (envelope) {
                envelope.markSpent()
            } else { throw new Error (`UNEXPECTED ${JSON.stringify(utxo)}`)}
        }
    }

    // mark the envelopes as spent
    markSpentEnvelopes(envelopes:any) {
        for( let x =0; x < envelopes.length; x++) {
            const envelope = envelopes[x]
            const envelopeFound = this.findutxo(envelope.utxo)
            if (envelope) {
                envelope.markSpent()
            } else { throw new Error (`UNEXPECTED ${JSON.stringify(envelope)}`)}
        }
    }

    // switch to selecting using envelopes
    selectUnspentEnvelopes(amount:number):Envelopes {
        // sort descending
        function compare(a:any, b:any) {          
            let comparison = 0;
            if (a.satoshis < b.satoshis) {
              comparison = 1;
            } else if (a.satoshis > b.satoshis) {
              comparison = -1;
            }
            return comparison;
          }
          this._token_envelopes.sort(compare)
          const result = new Envelopes()
          let sum = 0
          this._token_envelopes.forEach((env:any) => {
            if (sum < amount) {
                result.add(env)
                sum += env.satoshis
            }
          })
          return result
    }


}