// calls external apis
// we want to minimize calls these methods to eliminate dependency on external servers

// do not use @types/node-fetch because we may need to control credentials
const fetchFn: typeof fetch = require('node-fetch')
import { Address } from 'bsv'

async function portableFetch(
    input:RequestInfo, init?:RequestInit
) : Promise<Response> {
    return fetchFn(input, init)
}

export class IndexClient {

    swapEndianness (buffer:Buffer) {
        const result = Buffer.allocUnsafe(buffer.length)
        for (let i = 0, j = buffer.length - 1; i <= j; i++, j--) {
          result[i] = buffer[j]
          result[j] = buffer[i]
        }
        return result
      }

    async getBlockHeaderForMerkleRoot(merkleRoot:string) {
        const blockheaderurl = `http://li905-60.members.linode.com/${merkleRoot}`
        // const blockheaderurl = `http://45.56.103.60/${merkleRoot}`
        const res = await portableFetch(blockheaderurl)
        const block = await res.json()
        return block
    }

    // get utxos from WOC
    async getUnspents(address: string) {
        const utxourl = `https://api.whatsonchain.com/v1/bsv/main/address/${address}/unspent`
        const res = await portableFetch(utxourl)
        const utxos = await res.json()
        const adbuf = Address.fromString(address).hashBuf.toString('hex')
        
        // need script
        if (utxos) {
            for (let i:number=0; i<utxos.length; i++) {
                const utxo = utxos[i]
                utxo.script = '76a914' + adbuf + '88ac'
                //const tx = await this.getTx(utxo.tx_hash)
                //utxo.script = tx.vout[utxo.tx_pos].scriptPubKey.hex

            }
        }
        return utxos
        // [
        //     {
        //       height: 0,
        //       tx_pos: 1,
        //       tx_hash: '883ded7ef58f463087d0f6e6d53f0ed85bcf6c80fc201509c9d87f5e1a8c78e0',
        //       value: 7588
        //     }
        //   ]
    }

    //tx decoded into json
    async getTx(txid: string) {
        const url = `https://api.whatsonchain.com/v1/bsv/main/tx/hash/${txid}`
        const res = await portableFetch(url)
        return res.json()
        //return tx
    }

    // get raw tx hex
    async getTxRaw(txid: string) {
        const url = `https://api.whatsonchain.com/v1/bsv/main/tx/${txid}/hex`
        const res = await portableFetch(url)
        const tx = await res.text()
        return tx
    }

    // proof is WOC format, not TSC format, without index
    async getProof(txid: string) {
        const url =`https://api.whatsonchain.com/v1/bsv/main/tx/${txid}/proof`
        const res = await portableFetch(url)
        const tx = await res.json()
        return tx
    }

    async broadcast(raw: string) {
        return this.broadcastraw(raw)
    }

    // rawtx required, optional: url and token
    // dsCheck, merkleProof
    async broadcastmapi(mapistuff: any) {
        // https://github.com/bitcoin-sv-specs/brfc-merchantapi#2-submit-transaction
        const url_broadcast = `https://mapi.taal.com/mapi/tx`
        //const url_broadcast = `https://merchantapi.taal.com/mapi/tx`
        const mapisubmit:any = {
            "rawtx":        mapistuff.rawtx,
            "dsCheck" : mapistuff["dsCheck"] === undefined ? true :  mapistuff["dsCheck"],
            "merkleProof" : mapistuff["merkleProof"] === undefined ? false : mapistuff["merkleProof"],
            //  "callBackEncryption" : <parameter>
          }
          if (mapistuff.callBackUrl !== undefined) mapisubmit["callBackUrl"] = mapistuff.callBackUrl
          if (mapistuff.callBackToken !== undefined) mapisubmit["callBackToken"] = mapistuff.callBackToken
          if (mapistuff.merkleProof !== undefined) mapisubmit["merkleFormat"] = "TSC"

          console.log(mapisubmit)
          const res = await portableFetch(url_broadcast, {
            method: "POST",
            headers: {'Content-Type': 'application/json'},
            body: JSON.stringify(mapisubmit)
          })
          const broadcastResult = await res.json()
          return broadcastResult         
    }

    // if sent this way then you will not get ds check or merkleproof
    // use broadcastmapi for finer control
    async broadcastraw(raw: string) {
        return this.broadcastmapi({rawtx:raw, dsCheck:false, merkleProof:false})
        const url = `https://api.whatsonchain.com/v1/bsv/main/tx/raw`
        const data = {txhex: raw}
        const result = await portableFetch(url,
            {
                method:"POST",
                headers: {'Content-Type':'application/json'},
                body:JSON.stringify(data)
            }
        )
        return result.json()
    }

    async getFeeQuote () {
        //https://mapi.taal.com/mapi/feeQuote
        const mAPIHostEndpoint = 'mapi.taal.com'
        //const https = require('https')
        const feeQuoteOptions = {
            hostname: mAPIHostEndpoint,
            port: 443,
            path: '/mapi/feeQuote',
            method: 'GET'
        }

        const res = await portableFetch(`https://${feeQuoteOptions.hostname}${feeQuoteOptions.path}`)
        const resq = await res.json()
            
            // TODO: parse "payload" into var 
            // TODO: cache fee until specified expiry
            // apiVersion: "1.1.0"
            // timestamp:"2021-06-26T17:04:58.021Z"
            // expiryTime:"2021-06-26T17:14:58.021Z"
            // minerId:"03e92d3e5c3f7bd945dfbf48e7a99393b1bfb3f11f380ae30d286e7ff2aec5a270"
            // currentHighestBlockHash:"00000000000000000d7bd42359a02683fe0413cc8d625cd776abc5a1eea7f462"
            // currentHighestBlockHeight:693347
            // minerReputation:null
            // fees:[{\"id\":1,\"feeType\":\"standard\",\"miningFee\":{\"satoshis\":500,\"bytes\":1000},\"relayFee\":{\"satoshis\":250,\"bytes\":1000}},{\"id\":2,\"feeType\":\"data\",\"miningFee\":{\"satoshis\":500,\"bytes\":1000},\"relayFee\":{\"satoshis\":250,\"bytes\":1000}}]}",
        return resq        
    }

    // gets all mempool tx, use only for testing!
    async mempool() {
        const url = `https://api.whatsonchain.com/v1/bsv/main/mempool/raw`
        const f = await portableFetch(url)
        const res = await f.json()
        return res
    }

    async getBlockHeader(blockhash:string) {
        // This endpoint retrieves the block header details of a given hash.
        //GET https://api.whatsonchain.com/v1/bsv/<network>/block/<hash>/header
        const url = `https://api.whatsonchain.com/v1/bsv/main/block/${blockhash}/header`
        const f = await portableFetch(url)
        const res = await f.json()
        return res
    }

    async getBlockHeaderByNumber(blocknumber:string) {
        const url = `https://api.whatsonchain.com/v1/bsv/main/block/height/${blocknumber}`
        const f = await portableFetch(url)
        const res = await f.json()
        return res
    }

    async getBlockHeadersLast10() {
        const url = `https://api.whatsonchain.com/v1/bsv/main/block/headers`
        const f = await portableFetch(url)
        const res = await f.json()
        return res
    }

}